#include "SWFImportExport.h"
#include "SWF.h"
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include "readpng.h"
#include <zlib.h>

namespace SWF {

#define TMP_STRLEN 0xff
const char jpeg_header[] = { 0xff, 0xd9, 0xff, 0xd8, 0 };
		
void exportDefineBitsJPEG2( DefineBitsJPEG2 *tag, Context *ctx ) {
	int id = tag->getobjectID();
	FILE *fp = 0;
	
	Rest *rest = tag->getdata();
	unsigned char *data;
	int size;
	rest->getdata( &data, &size );
	if( size ) {

		char tmp[TMP_STRLEN];
		snprintf( tmp, TMP_STRLEN, "%i.jpg", id );
		fp = fopen( tmp, "wb" );
		if( fp ) {
			int ofs=strlen(jpeg_header); // FIXME: this is not really good, it can well fail. implement the table/data separation as per specs?
			if( fwrite( &data[ofs], 1, size-ofs, fp )!=size-ofs ) {
				goto fail;
			}
				
			fclose( fp );
		} else {
			fprintf(stderr,"WARNING: Could not export JPEG to %s\n", tmp );
		}
	}
	return;
	
fail:
	fprintf(stderr,"Could not export JPEG #%i\n", id );
	if( fp ) fclose(fp);
	return;
}

void importDefineBitsJPEG2( DefineBitsJPEG2 *tag, const char *filename, xmlNodePtr node, Context *ctx ) {
	unsigned char *data;
	int size, ofs;
	struct stat filestat;
		
	Rest *rest = tag->getdata();
	
	FILE *fp = fopen( filename, "rb" );
	if( !fp ) goto fail;
	
	if( stat( filename, &filestat ) ) goto fail;
	
	size = filestat.st_size;
	
	ofs = strlen(jpeg_header);
	
	data = new unsigned char[size+ofs];
	if( !data ) goto fail;
	memcpy( data, jpeg_header, ofs );
	if( fread( &data[ofs], 1, size, fp ) != size ) {
		fprintf(stderr,"WARNING: could not read enough (%i) bytes for jpeg %s\n", size, filename );
		goto fail;
	}
		
	rest->setdata( data, size+ofs );
	
	if( !ctx->quiet ) fprintf(stderr, "imported %s: %i bytes\n", filename, size );

	if( fp ) fclose(fp);
	if( data ) delete data;
	return;
	
fail:
	if( fp ) fclose(fp);
	if( data ) delete data;
	fprintf( stderr, "WARNING: could not import %s\n", filename );
	return;
}

// outLength should contain the allocated size,
// will be updated with the actual size
bool compress( unsigned char *inputBuffer, int inLength, unsigned char *outputBuffer, int *outLength ) {
	z_stream stream;
	int status, count;
	
	stream.avail_in = inLength;
	stream.next_in = inputBuffer;
	stream.next_out = outputBuffer;
	stream.zalloc = (alloc_func) NULL;
	stream.zfree = (free_func) NULL;
	stream.opaque = (voidpf) 0;
	stream.avail_out = *outLength;
	
	status = deflateInit( &stream, Z_BEST_COMPRESSION );
	if( status != Z_OK ) {
		fprintf( stderr, "Error compressing PNG: %s\n", stream.msg );
		return false;
	}
	
	while( true ) {
		if( stream.avail_in == 0 ) break;

		status = deflate( &stream, Z_NO_FLUSH );
		
		if( status != Z_OK ) {
			fprintf( stderr, "Error compressing PNG: %s\n", stream.msg );
			return false;
		}
	}

	do {
		status = deflate( &stream, Z_FINISH );
	} while( status == Z_OK );
	
	if( status != Z_STREAM_END ) {
		fprintf( stderr, "Error compressing PNG: %s\n", stream.msg );
		return false;
	}

	status = deflateEnd(&stream);

	if( status != Z_OK ) {
		fprintf( stderr, "Error compressing PNG: %s\n", stream.msg );
		return false;
	}
	
	*outLength -= stream.avail_out;
	return true;
}


void exportDefineBitsLossless2( DefineBitsLossless2 *tag, Context *ctx ) {
	// PNG Export not implemented
}

void importDefineBitsLossless2( DefineBitsLossless2 *tag, const char *filename, xmlNodePtr node, Context *ctx ) {
	FILE *fp = fopen( filename, "rb" );
	unsigned char *data, *compressed;
	unsigned long w, h, rowbytes;
	int channels;
	int compressed_size;
	if( !fp ) goto fail;
		
	if( readpng_init( fp, &w, &h ) ) goto fail;

	data = readpng_get_image( 2.2, &channels, &rowbytes );
	
	if( channels == 4 && rowbytes == (4*w) ) {
		int c;
		float a;
		for( int i=0; i<w*h*4; i+=4 ) {
			c = *((int*)&data[i]);
			data[i] = (c&0xff000000)>>24;
			a = data[i]/255.0;
			data[i+3] = (unsigned char)(((c&0x00ff0000)>>16)*a);
			data[i+2] = (unsigned char)(((c&0x0000ff00)>>8)*a);
			data[i+1] = (unsigned char)(((c&0x000000ff))*a);
		}
	} else if( channels == 3 && rowbytes == (3*w) ) {
		unsigned char *rgba = new unsigned char[ w*h*4 ];
		
		for( int i=0; i<w*h; i++ ) {
			rgba[i*4] = 0xff;
			rgba[(i*4)+1] = data[(i*3)+2];
			rgba[(i*4)+2] = data[(i*3)+1];
			rgba[(i*4)+3] = data[(i*3)];
		}
		data = rgba;
	} else {
		fprintf( stderr, "WARNING: can only import 24 or 32bit RGB(A) PNGs (%s has %i channels, rowstride %i)\n", filename, channels, rowbytes );
		goto fail;
	}
	
	compressed_size = w*h*4;
	compressed = new unsigned char[ compressed_size ];
	if( compress( data, w*h*4, compressed, &compressed_size ) ) {
		tag->setwidth( w );
		tag->setheight( h );
		tag->setformat( 5 );
		Rest *d = tag->getdata();
		d->setdata( compressed, compressed_size );
	}
	
	if( !ctx->quiet ) fprintf(stderr, "imported %s: %i bytes\n", filename, compressed_size );

	delete compressed;
	readpng_cleanup( true );
	if( fp ) fclose(fp);
	return;
	
fail:
	readpng_cleanup( true );
	if( fp ) fclose(fp);
	fprintf( stderr, "WARNING: could not import %s\n", filename );
	return;
}

}
